package de.popforge.cubicvr
{
	import de.popforge.cubicvr.config.Configuration;
	import de.popforge.cubicvr.gui.ControlPanel;
	import de.popforge.cubicvr.gui.Cursor;
	import de.popforge.cubicvr.smart3d.Camera;
	import de.popforge.cubicvr.smart3d.Cube;
	import de.popforge.cubicvr.smart3d.HotSpot;
	import de.popforge.cubicvr.smart3d.Polygon;
	import de.popforge.cubicvr.smart3d.Scanline;
	import de.popforge.cubicvr.smart3d.TextureLoader;
	import de.popforge.cubicvr.smart3d.Vector;
	
	import flash.display.Bitmap;
	import flash.display.Loader;
	import flash.display.PixelSnapping;
	import flash.display.Shape;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.external.ExternalInterface;
	import flash.filters.BlurFilter;
	import flash.geom.Point;
	import flash.net.URLRequest;
	import flash.utils.Timer;
	
	public class CubicVR extends Sprite
	{
		static private const MAX_ZOOM: Number = 8;
		static private const MIN_MOVE_OFFSET2: Number = 9;
		
		private var config: Configuration;
		private var cube: Cube;
		private var camera: Camera;
		private var texture: TextureLoader;
		private var control: ControlPanel;
		private var cursor: Cursor;
		private var output: Bitmap;
		
		private var splashLoader: Loader;
		
		private var mouseOver: Boolean = false;
		private var mouseDown: Boolean = false;
		private var mtx: int;
		private var mty: int;
		
		private var lowMoveQuality: Boolean;
		
		private var motionBlur: Boolean;
		
		//-- ZOOM
		private var zoomInKeyCode: uint;
		private var zoomOutKeyCode: uint;
		private var zoomIn: Boolean = false;
		private var zoomOut: Boolean = false;
		private var zoomQuad: Number = 0;
		private var zoomSpeed: Number;
		
		//-- TURN
		private var turnLeftKeyCode: uint;
		private var turnRightKeyCode: uint;
		private var turnSpeed: Number;
		
		//-- CONSTRAIN
		private var contrain_horizontal: Boolean;
		
		//-- HOTSPOTS
		private var hotSpotContainer: Sprite;
		private var hotSpotUnderMouse: HotSpot;
		
		//-- CACHE
		private var cache: Object;
		
		public function CubicVR( width: uint, height: uint )
		{
			config = new Configuration();
			config.onComplete = onConfigComplete;
			
			cursor = new Cursor();
			camera = new Camera( width, height );
			
			control = new ControlPanel();
			control.onZoom = zoom;
			control.resize( width );
			control.y = height;
			
			output = new Bitmap( camera.getOutput() );
			
			hotSpotContainer = new Sprite();
			hotSpotContainer.mouseEnabled = false;
			
			cache = new Object();
		}
		
		public function loadConfigXML( url: String ): void
		{
			config.loadXML( url );
		}
		
		public function requestCube( name: String ): void
		{
			if( cube != null )
			{
				cube.abortLoading();
				cube.removeHotSpots( hotSpotContainer );
			}
			
			hotSpotContainer.visible = false;
			
			if( cache[ name ] )
			{
				cube = cache[ name ];
				hotSpotContainer.visible = true;
			}
			else
			{
				cube = cache[ name ] = new Cube( this, config, name );
				
				cube.addEventListener( Event.COMPLETE, onCubeComplete );
				
				cube.load();
			}
			
			//-- reset
			camera.useMidMaps = cube.isTextureBlurred();
			camera.setZoom( 1 );
			camera.ax = 0;
			zoomQuad = 0;
		}
		
		private function onCubeComplete( event: Event ): void
		{
			hotSpotContainer.visible = true;
		}
		
		private function onConfigComplete(): void
		{
			if( config.getXML().splashscreen.@enable == 'true' )
			{
				splashLoader = new Loader();
				splashLoader.contentLoaderInfo.addEventListener( Event.COMPLETE, showSplashScreen );
				splashLoader.load( new URLRequest( config.getXML().splashscreen.@url ) );
			}
			else
			{
				startVR();
			}
		}
		
		private function showSplashScreen( event: Event ): void
		{
			addChild( splashLoader );
			
			var timeout: Timer = new Timer( parseInt( config.getXML().splashscreen.@timeout ), 1 );
			
			timeout.addEventListener( TimerEvent.TIMER, removeSplashScreen );
			timeout.start();
		}
		
		private function removeSplashScreen( event: Event ): void
		{
			removeChild( splashLoader );
			
			startVR();
		}		
		
		private function startVR(): void
		{
			addChild( control );
			
			var browserBoost: Timer = new Timer( 20 );
			browserBoost.addEventListener( TimerEvent.TIMER, onEnterFrame );
			//browserBoost.start();
			
			stage.addEventListener( Event.ENTER_FRAME, onEnterFrame );
			stage.addEventListener( MouseEvent.MOUSE_DOWN, onMouseDown );
			stage.addEventListener( MouseEvent.MOUSE_UP, onMouseUp );
			stage.addEventListener( MouseEvent.MOUSE_OUT, onMouseOut );
			stage.addEventListener( MouseEvent.MOUSE_OVER, onMouseOver );
			
			Keys.initStage( stage );
			
			addChild( output );
			addChild( hotSpotContainer );
			addChild( cursor );
			
			//-- GLOBAL XML CONFIGURATION
			camera.eyeAngle = parseInt( config.getXML().env.param.(@name == 'fov' ).attribute( 'value' ) );
			zoomInKeyCode = parseInt( config.getXML().env.param.(@name == 'zoomIn' ).attribute( 'value' ) );
			zoomOutKeyCode = parseInt( config.getXML().env.param.(@name == 'zoomOut' ).attribute( 'value' ) );
			zoomSpeed = parseFloat( config.getXML().env.param.(@name == 'zoomSpeed' ).attribute( 'value' ) );
			contrain_horizontal = config.getXML().env.param.(@name == 'constrain-horizontal' ).attribute( 'value' ) == 'true';
			motionBlur = config.getXML().env.param.(@name == 'motion-blur' ).attribute( 'value' ) == 'true';
			lowMoveQuality = config.getXML().env.param.(@name == 'low-resolution-while-move' ).attribute( 'value' ) == 'true';
			turnLeftKeyCode = parseInt( config.getXML().env.param.(@name == 'turnLeft' ).attribute( 'value' ) );
			turnRightKeyCode = parseInt( config.getXML().env.param.(@name == 'turnRight' ).attribute( 'value' ) );
			turnSpeed = parseFloat( config.getXML().env.param.(@name == 'turnSpeed' ).attribute( 'value' ) );
			
			requestCube( config.getXML().vr.@startcube );
			
			cursor.show( 'orbit' );
		}
		
		private function zoom( delta: int ): void
		{
			zoomIn = delta > 0;
			zoomOut = delta < 0;
		}
		
		private function onEnterFrame( event: Event ): void
		{
			var moved: Boolean = false;
			
			if( Keys.isDown( zoomInKeyCode ) || zoomIn )
			{
				zoomQuad += zoomSpeed;
				if( zoomQuad > MAX_ZOOM ) zoomQuad = MAX_ZOOM;
				moved = true;
			}
			else if( Keys.isDown( zoomOutKeyCode ) || zoomOut )
			{
				zoomQuad -= zoomSpeed;
				if( zoomQuad < 0 ) zoomQuad = 0;
				moved = true;
			}
			
			if( Keys.isDown( turnLeftKeyCode ) )
			{
				camera.ay -= turnSpeed;
				moved = true;
			}
			
			if( Keys.isDown( turnRightKeyCode ) )
			{
				camera.ay += turnSpeed;
				moved = true;
			}
			
			if( mouseDown )
			{
				var dx: Number = mouseX - mtx;
				var dy: Number = mouseY - mty;
				
				if( dx * dx + dy * dy > MIN_MOVE_OFFSET2 )
				{
					if( lowMoveQuality ) camera.setQuality( Camera.LOW );
					
					camera.ay += dx / 2000;
					
					if( contrain_horizontal )
					{
						cursor.show( 'cursor_move_' + ( dx < 0 ? 0 : 4 ) );
					}
					else
					{
						camera.ax -= dy / 2000;
					
						if( camera.ax < -Math.PI / 2 )
						{
							camera.ax = -Math.PI / 2;
						}
						else if( camera.ax > Math.PI / 2 )
						{
							camera.ax = Math.PI / 2;
						}
						
						var angle: Number = ( Math.atan2( -dy, -dx ) + Math.PI / 8 ) / Math.PI * 4;
						
						if( angle < 0 ) angle += 8;
						
						cursor.show( 'cursor_move_' + Math.floor( angle ) );
					}
				}
				else
				{
					cursor.show( 'move' );
				}
				
				moved = true;
			}
			
			camera.updateMatrix();
			
			camera.setZoom( 1 + zoomQuad * zoomQuad );
			
			//-- RENDERING
			
			camera.render( cube.getPolygons(), 0 );
			
			if( motionBlur )
			{
				output.filters = [ new BlurFilter( Math.abs( dx ) >> 4, Math.abs( dy ) >> 4, 2 ) ];
			}
			
			hotSpotUnderMouse = cube.getNearestHotSpot( camera, mouseX, mouseY );
			
			cube.showHotSpots( camera, hotSpotContainer );
			
			if( hotSpotUnderMouse != null )
			{
				cursor.show( 'link' );
			}
			else if( !moved )
			{
				cursor.show( null );
			}
		}
		
		private function onMouseDown( event: MouseEvent ): void
		{
			if( mouseOver )
			{
				if( hotSpotUnderMouse )
				{
					hotSpotUnderMouse.doAction();
				}
				else
				{
					cursor.show( 'move' );
				}
				
				mouseDown = true;
				
				mtx = mouseX;
				mty = mouseY;
			}
		}
		
		private function onMouseUp( event: MouseEvent ): void
		{
			mouseDown = false;
		
			cursor.show( 'orbit' );
			
			if( lowMoveQuality )
			{
				camera.setQuality( Camera.MEDIUM );
				camera.render( cube.getPolygons() );
			}
		}
		
		private function onMouseOut( event: MouseEvent ): void
		{
			if( event.target == this )
			{
				cursor.show( null );
				mouseOver = false;
			}
		}
		
		private function onMouseOver( event: MouseEvent ): void
		{
			if( event.target == this )
			{
				cursor.show( 'orbit' );
				mouseOver = true;
			}
		}
	}
}