package de.popforge.cubicvr.smart3d
{
	import de.popforge.cubicvr.CubicVR;
	import de.popforge.cubicvr.config.Configuration;
	
	import flash.display.BitmapData;
	import flash.display.DisplayObjectContainer;
	import flash.geom.Point;
	import flash.events.EventDispatcher;
	import flash.events.Event;
	
	public class Cube extends EventDispatcher
	{
		private var app: CubicVR;
		
		private var name: String;
		
		private var polygons: Array;
		private var loaders: Array;
		private var textureBlur: Boolean;
		
		private var hotSpots: Array;
		
		public function Cube( app: CubicVR, config: Configuration, name: String )
		{
			this.app = app;
			
			this.name = name;
			
			init( config, name );
		}
		
		public function getName(): String
		{
			return name;
		}
		
		public function getPolygons(): Array
		{
			return polygons.slice( 0 );
		}
		
		public function isTextureBlurred(): Boolean
		{
			return textureBlur;
		}
		
		public function getNearestHotSpot( camera: Camera, sx: Number, sy: Number ): HotSpot
		{
			var dx: Number;
			var dy: Number;
			var dd: Number;
			
			var nearestHotSpot: HotSpot = null;
			var nearestDistance: Number = Number.POSITIVE_INFINITY;
			
			for each( var hotSpot: HotSpot in hotSpots )
			{
				camera.inverseMatrix.transformVertex( Vertex( hotSpot ) );
				camera.project( Vertex( hotSpot ) );
				
				dx = hotSpot.sx - sx;
				dy = hotSpot.sy - sy;
				dd = Math.sqrt( dx * dx + dy * dy );
				
				if( dd < hotSpot.sensitivRadius )
				{
					if( dd < nearestDistance )
					{
						nearestHotSpot = hotSpot;
						nearestDistance = dd;
					}
				}
			}
			
			return nearestHotSpot;
		}
		
		public function showHotSpots( camera: Camera, displayObject: DisplayObjectContainer ): void
		{
			for each( var hotSpot: HotSpot in hotSpots )
			{
				camera.inverseMatrix.transformVertex( Vertex( hotSpot ) );
				if( camera.project( Vertex( hotSpot ) ) )
				{
					hotSpot.show( displayObject );
				}
				else
				{
					hotSpot.hide( displayObject );
				}
			}
		}
		
		public function removeHotSpots( displayObject: DisplayObjectContainer ): void
		{
			for each( var hotSpot: HotSpot in hotSpots )
			{
				hotSpot.hide( displayObject );
			}
		}
		
		public function abortLoading(): void
		{
			for each( var loader: TextureLoader in loaders )
			{
				loader.abortLoading();
			}
		}
		
		private function init( config: Configuration, name: String ): void
		{
			polygons = new Array();
			hotSpots = new Array();
			loaders = new Array();
			
			var textureXML: XMLList;
			var hotspotXMLList: XMLList;
			var hotspotXML: XML;
			var sideXML: XMLList;
			
			var texture: BitmapData;
			var polygon: Polygon;
			var hotSpot: HotSpot;
			
			var xmlList: XMLList = config.getXML().vr.cube.(@name == name );
			
			textureBlur = xmlList.@textureBlur == 'true';
			
			var textureSize: int = 1 << ( parseInt( xmlList.@textureSize ) + 7 );
			
			//-- DOWN
			sideXML = xmlList.side.(@align == 'down');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					1,
					1 - parseFloat( hotspotXML.@u ) / textureSize * 2,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			var loader: TextureLoader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex(  1,  1,  1, 1, 0 ),
					new Vertex( -1,  1,  1, 0, 0 ),
					new Vertex( -1,  1, -1, 0, 1 ),
					new Vertex(  1,  1, -1, 1, 1 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
			
			//-- UP
			sideXML = xmlList.side.(@align == 'up');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					-1,
					parseFloat( hotspotXML.@u ) / textureSize * 2 - 1,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			loader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex( -1, -1,  1, 0, 1 ),
					new Vertex(  1, -1,  1, 1, 1 ),
					new Vertex(  1, -1, -1, 1, 0 ),
					new Vertex( -1, -1, -1, 0, 0 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
			
			//-- LEFT
			sideXML = xmlList.side.(@align == 'left');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					-1,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					parseFloat( hotspotXML.@u ) / textureSize * 2 - 1,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			loader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex( -1,  1, -1, 0, 1 ),
					new Vertex( -1,  1,  1, 1, 1 ),
					new Vertex( -1, -1,  1, 1, 0 ),
					new Vertex( -1, -1, -1, 0, 0 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
			
			//-- BACK
			sideXML = xmlList.side.(@align == 'back');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					1 - parseFloat( hotspotXML.@u ) / textureSize * 2,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					-1,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			loader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex(  1,  1, -1, 0, 1 ),
					new Vertex( -1,  1, -1, 1, 1 ),
					new Vertex( -1, -1, -1, 1, 0 ),
					new Vertex(  1, -1, -1, 0, 0 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
			
			//-- RIGHT
			sideXML = xmlList.side.(@align == 'right');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					1,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					1 - parseFloat( hotspotXML.@u ) / textureSize * 2,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			loader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex(  1,  1,  1, 0, 1 ),
					new Vertex(  1,  1, -1, 1, 1 ),
					new Vertex(  1, -1, -1, 1, 0 ),
					new Vertex(  1, -1,  1, 0, 0 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
			
			//-- FRONT
			sideXML = xmlList.side.(@align == 'front');
			
			textureXML = sideXML.texture;
			hotspotXMLList = sideXML.hotspot;
			
			for each( hotspotXML in hotspotXMLList )
			{
				hotSpot = new HotSpot
				(
					app,
					parseFloat( hotspotXML.@u ) / textureSize * 2 - 1,
					parseFloat( hotspotXML.@v ) / textureSize * 2 - 1,
					1,
					parseFloat( hotspotXML.@sensitivRadius ),
					hotspotXML.@action,
					hotspotXML.@target
				);
				
				hotSpots.push( hotSpot );
			}
			
			loader = new TextureLoader( textureSize, textureXML.@url, textureBlur );
			
			polygon = new Polygon
			(
				[
					new Vertex( -1,  1,  1, 0, 1 ),
					new Vertex(  1,  1,  1, 1, 1 ),
					new Vertex(  1, -1,  1, 1, 0 ),
					new Vertex( -1, -1,  1, 0, 0 )
				]
				,loader.getTextureNormal()
				,loader.getTextureBlurred()
			);
			
			loaders.push( loader );
			polygons.push( polygon );
		}
		
		public function load(): void
		{
			var loader: TextureLoader;
			var prev: TextureLoader;
			
			while( loader = TextureLoader( loaders.pop() ) )
			{
				if( prev == null )
				{
					loader.load();
				}
				else
				{
					prev.onLoad = loader.load;
				}
				
				prev = loader;
			}
			
			prev.onLoad = onCubeComplete;
		}
		
		private function onCubeComplete(): void
		{
			dispatchEvent( new Event( Event.COMPLETE ) );
		}
	}
}