package de.popforge.cubicvr.smart3d
{
	import flash.display.BitmapData;
	import flash.geom.Rectangle;
	
	public class Scanline
	{
		private var output: BitmapData;
		
		private var dizdx: Number;
		private var duizdx: Number;
		private var dvizdx: Number;
		private var dizdy: Number;
		private var duizdy: Number;
		private var dvizdy: Number;

		private var xa: Number;
		private var xb: Number;
		private var iza: Number;
		private var uiza: Number;
		private var viza: Number;

		private var dxdya: Number;
		private var dxdyb: Number;
		private var dizdya: Number;
		private var duizdya: Number;
		private var dvizdya: Number;

		private var texture: BitmapData;
		private var rw: int;
		private var rh: int;

		public function Scanline( width: int, height: int )
		{
			output = new BitmapData( width, height, false, 0 );
		}

		public function clear( background: uint = 0 ): void
		{
			//output.fillRect( output.rect, background );
		}
		
		public function getOutput(): BitmapData
		{
			return output;
		}
		
		public function drawPolygon( poly: Polygon ): void
		{
			/*
				tesselate polygons over 3 vertices recursiv
				[fast and ugly solution]
			*/
			var pvert: Array = poly.cvert.slice();
			
			if( pvert.length > 3 )
			{
				pvert = pvert.slice( 0, 3 );
				poly.cvert.splice( 1, 1 );
				
				drawPolygon( poly );
			}
			
			//-- Native AS3 Sort ( Sort the vertices in ascending Y order )		
			var sySort: Array = pvert.sortOn( 'sy', Array.RETURNINDEXEDARRAY | Array.NUMERIC );

			var v0: Vertex = pvert[sySort[0]];
			var v1: Vertex = pvert[sySort[1]];
			var v2: Vertex = pvert[sySort[2]];
			
			//-- Shift XY coordinate system (+0.5, +0.5) to match the subpixeling technique
			
			var x1: Number = v0.sx + .5;
			var y1: Number = v0.sy + .5;
			var x2: Number = v1.sx + .5;
			var y2: Number = v1.sy + .5;
			var x3: Number = v2.sx + .5;
			var y3: Number = v2.sy + .5;
			
			//-- Calculate alternative 1/Z, U/Z and V/Z values which will be
			//-- interpolated
			
			var iz1: Number = v0.sz;
			var iz2: Number = v1.sz;
			var iz3: Number = v2.sz;
			
			texture = poly.texture;
			
			var tw: int = texture.width;
			var th: int = texture.height;
			
			rw = tw - 1;
			rh = th - 1;

			var uiz1: Number = v0.u * iz1 * tw;
			var viz1: Number = v0.v * iz1 * th;
			var uiz2: Number = v1.u * iz2 * tw;
			var viz2: Number = v1.v * iz2 * th;
			var uiz3: Number = v2.u * iz3 * tw;
			var viz3: Number = v2.v * iz3 * th;
			
			var y1i: int = int( y1 );
			var y2i: int = int( y2 );
			var y3i: int = int( y3 );
		
			// Skip poly if it's too thin to cover any pixels at all
		
			// Calculate horizontal and vertical increments for UV axes (these
			// calcs are certainly not optimal, although they're stable
			// (handles any dy being 0)
		
			var denom: Number = ((x3 - x1) * (y2 - y1) - (x2 - x1) * (y3 - y1));

			var dy: Number;

			dizdx = ((iz3 - iz1) * (y2 - y1) - (iz2 - iz1) * (y3 - y1)) / denom;
			duizdx = ((uiz3 - uiz1) * (y2 - y1) - (uiz2 - uiz1) * (y3 - y1)) / denom;
			dvizdx = ((viz3 - viz1) * (y2 - y1) - (viz2 - viz1) * (y3 - y1)) / denom;
			dizdy = ((iz2 - iz1) * (x3 - x1) - (iz3 - iz1) * (x2 - x1)) / denom;
			duizdy = ((uiz2 - uiz1) * (x3 - x1) - (uiz3 - uiz1) * (x2 - x1)) / denom;
			dvizdy = ((viz2 - viz1) * (x3 - x1) - (viz3 - viz1) * (x2 - x1)) / denom;
		
			// Calculate X-slopes along the edges
		
			var dxdy1: Number = ( x2 - x1 ) / ( y2 - y1 );
			var dxdy2: Number = ( x3 - x1 ) / ( y3 - y1 );
			var dxdy3: Number = ( x3 - x2 ) / ( y3 - y2 );
		
			// Determine which side of the poly the longer edge is on
		
			var side: Boolean = dxdy2 > dxdy1;
		
			if( y1 == y2 )
			{
				side = x1 > x2;
			}
			if( y2 == y3 )
			{
				side = x3 > x2;
			}
		
			if( !side )	// Longer edge is on the left side
			{
				// Calculate slopes along left edge

				dxdya = dxdy2;
				dizdya = dxdy2 * dizdx + dizdy;
				duizdya = dxdy2 * duizdx + duizdy;
				dvizdya = dxdy2 * dvizdx + dvizdy;
		
				// Perform subpixel pre-stepping along left edge
		
				dy = 1 - ( y1 - y1i );
				
				xa = x1 + dy * dxdya;
				iza = iz1 + dy * dizdya;
				uiza = uiz1 + dy * duizdya;
				viza = viz1 + dy * dvizdya;
		
				if (y1i < y2i)	// Draw upper segment if possibly visible
				{
					// Set right edge X-slope and perform subpixel pre-stepping
		
					xb = x1 + dy * dxdy1;
					dxdyb = dxdy1;
		
					drawSubTri( y1i, y2i );
				}
					

				if (y2i < y3i)	// Draw lower segment if possibly visible
				{
					// Set right edge X-slope and perform subpixel pre-stepping
		
					xb = x2 + (1 - (y2 - y2i)) * dxdy3;
					dxdyb = dxdy3;
		
					drawSubTri( y2i, y3i );
				}
			}
			else	// Longer edge is on the right side
			{
				// Set right edge X-slope and perform subpixel pre-stepping
		
				dxdyb = dxdy2;
				dy = 1 - (y1 - y1i);
				xb = x1 + dy * dxdyb;
		
				if( y1i < y2i )	// Draw upper segment if possibly visible
				{
					// Set slopes along left edge and perform subpixel pre-stepping

					dxdya = dxdy1;
					dizdya = dxdy1 * dizdx + dizdy;
					duizdya = dxdy1 * duizdx + duizdy;
					dvizdya = dxdy1 * dvizdx + dvizdy;
					xa = x1 + dy * dxdya;
					iza = iz1 + dy * dizdya;
					uiza = uiz1 + dy * duizdya;
					viza = viz1 + dy * dvizdya;

					drawSubTri( y1i, y2i );
				}
				if( y2i < y3i )	// Draw lower segment if possibly visible
				{
					// Set slopes along left edge and perform subpixel pre-stepping

					dxdya = dxdy3;
					dizdya = dxdy3 * dizdx + dizdy;
					duizdya = dxdy3 * duizdx + duizdy;
					dvizdya = dxdy3 * dvizdx + dvizdy;

					dy = 1 - ( y2 - y2i );
					xa = x2 + dy * dxdya;
					iza = iz2 + dy * dizdya;
					uiza = uiz2 + dy * duizdya;
					viza = viz2 + dy * dvizdya;

					drawSubTri( y2i, y3i );
				}
			}
		}

		private function drawSubTri( y1: int, y2: int ): void
		{
			var x1: int;
			var x2: int;
			var dx: Number;
			var iz: Number;
			var uiz: Number;
			var viz: Number;
			
			while ( y1 < y2 )
			{
				x1 = xa;
				x2 = xb;

				dx = 1 - ( xa - x1 );
				iz =   iza + dx * dizdx;
				uiz = uiza + dx * duizdx;
				viz = viza + dx * dvizdx;
				
				while( x1 < x2 )
				{
					output.setPixel( x1, y1, texture.getPixel( uiz / iz & rw, viz / iz & rh ) );

					iz += dizdx;
					uiz += duizdx;
					viz += dvizdx;
					
					x1++;
				}

				xa += dxdya;
				xb += dxdyb;
				iza += dizdya;
				uiza += duizdya;
				viza += dvizdya;

				y1++;
			}
		}
	}
}