package de.popforge.cubicvr.smart3d
{
	import flash.display.BitmapData;
	import flash.display.Graphics;
	import flash.display.Loader;
	import flash.display.Shape;
	import flash.events.Event;
	import flash.events.HTTPStatusEvent;
	import flash.events.IOErrorEvent;
	import flash.events.ProgressEvent;
	import flash.events.TimerEvent;
	import flash.geom.Matrix;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;
	import flash.utils.Timer;
	import flash.filters.ConvolutionFilter;
	
	public class TextureLoader
	{
		static public const ORIGIN: Point = new Point();
		
		public var onLoad: Function;
		
		private var size: uint;
		private var url: String;
		private var blur: Boolean;
		
		private var loader: Loader;
		private var textureNormal: BitmapData;
		private var textureBlurred: BitmapData;
		
		private var blurClip: Rectangle;
		private var blurThread: Timer;
		private var blurFilter: ConvolutionFilter;
		
		private var loadedBlocks: int;
		private var displayedBlocks: int;
		
		public function TextureLoader( size: uint, url: String, blur: Boolean )
		{
			this.size = size;
			this.url = url;
			this.blur = blur;
			
			textureNormal = new BitmapData( size, size, false, 0 );
			textureBlurred = new BitmapData( size, size, false, 0 );
			
			createLoadScreen();
		}
		
		public function getTextureNormal(): BitmapData
		{
			return textureNormal;
		}
		
		public function getTextureBlurred(): BitmapData
		{
			return textureBlurred;
		}
		
		public function abortLoading(): void
		{
			if( loader != null )
			{
				loader.close();
			}
		}
		
		public function load(): void
		{
			loader = new Loader();
			loader.contentLoaderInfo.addEventListener( Event.COMPLETE, onLoadHandler );
			loader.contentLoaderInfo.addEventListener( IOErrorEvent.IO_ERROR, onIOErrorHandler );
			loader.contentLoaderInfo.addEventListener( HTTPStatusEvent.HTTP_STATUS, onHTTPStatusHandler );
			loader.contentLoaderInfo.addEventListener( ProgressEvent.PROGRESS, onProgressHandler );
			loader.load( new URLRequest( url ) );
			
			displayedBlocks = loadedBlocks = 0;
		}
		
		private function onIOErrorHandler( event: IOErrorEvent ): void
		{
			trace( event );
		}
		
		private function onHTTPStatusHandler( event: HTTPStatusEvent ): void
		{
			trace( event );
		}
		
		private function onLoadHandler( event: Event ): void
		{
			textureNormal.fillRect( textureNormal.rect, 0 );
			textureNormal.draw( loader.content );
			
			loader.contentLoaderInfo.removeEventListener( Event.COMPLETE, onLoadHandler );
			loader.contentLoaderInfo.removeEventListener( ProgressEvent.PROGRESS, onProgressHandler );
			
			loader = null;
			
			if( blur )
			{
				var m: Array =
				[
					1, 1, 1,
					1, 1, 1,
					1, 1, 1
				];
				
				blurFilter = new ConvolutionFilter( 3, 3, m, 9, 0, true, true );
				blurClip = new Rectangle( 0, 0, textureNormal.width, 32 );
				blurThread = new Timer( 2, 0 );
				blurThread.addEventListener( TimerEvent.TIMER, blurring );
				blurThread.start();				
			}
			else if( onLoad != null ) onLoad();
		}
		
		private function blurring( event: TimerEvent ): void
		{
			textureBlurred.applyFilter( textureNormal, blurClip, blurClip.topLeft, blurFilter );
			
			if( ( blurClip.y += 32 ) == textureNormal.height )
			{
				blurThread.removeEventListener( TimerEvent.TIMER, blurring );
				
				if( onLoad != null ) onLoad();
			}
		}
		
		private function onProgressHandler( event: ProgressEvent ): void
		{
			loadedBlocks = Math.floor( event.bytesLoaded / event.bytesTotal * 256 );
			
			while( loadedBlocks > displayedBlocks )
			{
				var cs: uint = size >> 5;
				var os: uint = Math.log( size ) / Math.log( 2 ) - 4;
				
				var x: int = displayedBlocks & 15;
				var y: int = displayedBlocks >> 4;
				
				var c: Rectangle = new Rectangle( ( x << os ) + cs/2, ( y << os ) + cs/2, cs, cs );
				
				textureNormal.fillRect( c, 0xff3d3d3d );
				textureBlurred.fillRect( c, 0xff3d3d3d );
				
				displayedBlocks++;
			}
		}
		
		private function createLoadScreen(): void
		{
			//-- PATTERN
			var psize: uint = size >> 4;
			var outline: uint = size >> 8;
			
			var pattern: BitmapData = new BitmapData( psize, psize, false, 0x454545 );
			
			var rect: Rectangle = new Rectangle( 0, 0, psize, outline );
			
			pattern.fillRect( rect, 0x333333 );
			rect.y = psize - outline;
			pattern.fillRect( rect, 0x333333 );
			rect.width = outline; rect.y = 0;
			rect.height = psize;
			pattern.fillRect( rect, 0x333333 );
			rect.x = psize - outline;
			pattern.fillRect( rect, 0x333333 );
			
			var shape: Shape = new Shape();
			var g: Graphics = shape.graphics;
			
			g.beginBitmapFill( pattern, new Matrix(), true, false );
			g.drawRect( 0, 0, size, size );
			textureNormal.draw( shape );
			textureBlurred.draw( shape );
		}
	}
}